from typing import Any

def tert(condition: bool, message: str) -> None:
    if condition:
        return
    raise TypeError(message)

def typert(thing: Any, cls: type|list[type], name: str) -> None:
    if type(cls) is list:
        tert(type(thing) in cls,
             f"{name} must be instance of one of {[c.__name__ for c in cls]}")
    else:
        tert(isinstance(thing, cls), f"{name} must be instance of {cls.__name__}")

def vert(condition: bool, message: str) -> None:
    if condition:
        return
    raise ValueError(message)



from __future__ import annotations
from .errors import tert, typert, vert
from dataclasses import dataclass, field
from math import ceil, log
from random import randint
from typing import Callable


alphanumerics = [
    *[chr(i) for i in range(48, 58)],
    *[chr(i) for i in range(65, 91)],
    *[chr(i) for i in range(97, 123)],
]


def random_str(size: int) -> str:
    """Returns a str of random alphanumeric chars."""
    l = len(alphanumerics)-1
    return "".join([
        alphanumerics[randint(0, l)] for _ in range(size)
    ])


@dataclass
class Gene:
    """Represents a gene comprised of coding bases with a name of some type."""
    name: str = field(default_factory=lambda: random_str(4))
    bases: list[int|float|str] = field(default_factory=list)

    def copy(self) -> Gene:
        """Returns an exact copy of the Gene."""
        return Gene(name=self.name, bases=[*self.bases])

    def insert(self, index: int = None, base: int|float|str = None) -> Gene:
        """Inserts the base at the index. If index is None, the base is
            inserted at a random index. If base is None, adds a random
            int base. Returns self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.bases)-1)
        base = base if base is not None else randint(0, max(self.bases))
        typert(index, int, "index")
        vert(index < len(self.bases), "index must be < len(bases)")
        typert(base, [int, float, str], "base")
        self.bases.insert(index, base)
        return self

    def append(self, base: int|float|str = None) -> Gene:
        """Adds a base to the end of the gene. If base is None, adds a
            random int base. Returns self for chaining operations.
        """
        base = base if base is not None else randint(0, max(self.bases))
        typert(base, [int, float, str], "base")
        self.bases.append(base)
        return self

    def insert_sequence(self, index: int = None,
                        sequence: list[int|float|str] = None) -> Gene:
        """Inserts the sequence at the index. If index is None, the
            sequence is inserted at a random index. If sequence is None,
            a random sequence in size between 1 and the current len of
            the gene bases will be inserted.
        """
        index = index if index is not None else randint(0, len(self.bases)-1)
        if sequence is None:
            size = randint(1, len(self.bases))
            max_base = max(self.bases)
            sequence = [randint(0, max_base) for _ in range(size)]
        typert(index, int, "index")
        vert(index < len(self.bases), "index must be < len(bases)")
        tert(type(sequence) is list, "sequence must be list[int|float|str]")
        tert(all(type(s) in (int, float, str) for s in sequence),
             "sequence must be list[int|float|str]")
        self.bases = [*self.bases[:index], *sequence, *self.bases[index:]]
        return self

    def delete(self, index: int = None) -> Gene:
        """Deletes the base at the index. If index is None, a random
            base is deleted. Returns self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.bases)-1)
        typert(index, int, "index")
        vert(index < len(self.bases), "index must be < len(bases)")
        del self.bases[index]
        return self

    def delete_sequence(self, index: int = None, size: int = None) -> Gene:
        """Deletes size bases beggining at the index. If index is None,
            a random index is used. If size is None, a random size is
            used. Returns self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.bases)-1)
        size = size if size is not None else randint(1, len(self.bases)-index)
        typert(index, int, "index")
        vert(index < len(self.bases), "index must be < len(bases)")
        typert(size, int, "size")
        vert(size > 0, "size must be > 0")
        del self.bases[index:index+size]
        return self

    def substitute(self, index: int = None, base: int|float|str = None) -> Gene:
        """Substitutes the base at the index with the given base. If
            index is None, a random index will be used. If base is None,
            a random int base will be used. Returns self for chaining
            operations.
        """
        index = index if index is not None else randint(0, len(self.bases)-1)
        base = base if base is not None else randint(0, max(self.bases))
        typert(index, int, "index")
        vert(index < len(self.bases), "index must be < len(bases)")
        typert(base, [int, float, str], "base")
        self.bases[index] = base
        return self

    def recombine(self, other: Gene, indices: list[int] = None) -> Gene:
        """Recombines with another gene at the given indexes. If indices
            is None, between 1 and ceil(log(len(self.bases))) random
            indices will be chosen. Returns the resultant Gene.
        """
        typert(other, Gene, "other")
        vert(len(other.bases) > 0, "other must have bases")
        max_size = min(len(self.bases), len(other.bases))
        max_swaps = ceil(log(max_size)) or 1
        tert(indices is None or type(indices) is list,
             "indices must be list[int] or None")
        if type(indices) is list:
            tert(all(type(i) is int for i in indices),
                 "indices must be list[int] or None")
            vert(len(indices) <= max_size,
                 f"can have at most {max_size} indices")
        else:
            swaps = randint(1, max_swaps)
            indices = list(set([randint(0, max_size-1) for _ in range(swaps)]))
            indices.sort()

        name = self.name
        if self.name != other.name:
            name_size = min(len(self.name), len(other.name))
            name_swap = randint(1, name_size-1)
            name = self.name[:name_swap] + other.name[name_swap:]

        bases = [*self.bases]
        swapped = False
        for i in indices:
            bases[i:] = self.bases[i:] if swapped else other.bases[i:]
            swapped = not swapped

        return Gene(name=name, bases=bases)

    @classmethod
    def make(cls, n_bases: int, name: str = None, *, max_base_size: int = 10,
             base_factory: Callable[[None], int|float|str] = None,
             factory_args: list = [], factory_kwargs: dict = {},) -> Gene:
        """Makes and returns a randomized Gene."""
        if base_factory is not None and callable(base_factory):
            bases = [
                base_factory(*factory_args, **factory_kwargs)
                for _ in range(n_bases)
            ]
        else:
            bases = [randint(0, max_base_size) for _ in range(n_bases)]
        if name:
            return cls(name=name, bases=bases)
        return cls(bases=bases)

    def to_dict(self) -> dict:
        """Serialize the Gene to a dict."""
        return {self.name: [*self.bases]}

    @classmethod
    def from_dict(cls, data: dict) -> Gene:
        """Deserialize a Gene from a dict."""
        for name, bases in data.items():
            return cls(name=name, bases=bases)

    def __hash__(self) -> int:
        """Make Gene hashable."""
        return hash((self.name, tuple(self.bases)))


@dataclass
class Allele:
    name: str = field(default_factory=lambda: random_str(3))
    genes: list[Gene] = field(default_factory=list)

    def copy(self) -> Allele:
        """Returns an exact copy of the Allele, copying the underlying
            Genes as well.
        """
        return Allele(name=self.name, genes=[g.copy() for g in self.genes])

    def insert(self, index: int = None, gene: Gene = None, *,
               n_bases: int = None, max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Allele:
        """Inserts the gene at the index. If index is None, the gene is
            inserted at a random index. If gene is None, adds a random
            gene using Gene.make, passing the kwargs. Returns self for
            chaining operations.
        """
        index = index if index is not None else randint(0, len(self.genes)-1)
        typert(index, int, "index")
        vert(0 <= index <= len(self.genes), "index out of range")

        if gene is None:
            if n_bases is None:
                n_bases = randint(0, max(len(g.bases) for g in self.genes))
            typert(n_bases, int, "n_bases")
            gene = Gene.make(
                n_bases, max_base_size=max_base_size, base_factory=base_factory,
                factory_args=factory_args,factory_kwargs=factory_kwargs
            )
        typert(gene, Gene, "gene")

        self.genes.insert(index, gene)
        return self

    def append(self, gene: Gene = None, /, *, n_bases: int = None,
               max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Allele:
        """Adds a gene to the end of the allele. If gene is None, adds a
            random Gene using Gene.make and passing kwargs. Returns self
            for chaining operations.
        """
        return self.insert(
            len(self.genes), gene, n_bases=n_bases, max_base_size=max_base_size,
            base_factory=base_factory, factory_args=factory_args,
            factory_kwargs=factory_kwargs
        )

    def duplicate(self, index: int = None) -> Allele:
        """Duplicates the Gene at the index. If index is None, the Gene
            at a random index is duplicated. Returns self for chaining
            operations.
        """
        index = index if index is not None else randint(0, len(self.genes)-1)
        typert(index, int, "index")
        vert(0 <= index < len(self.genes), "index out of range")
        return self.insert(index, self.genes[index].copy())

    def delete(self, index: int = None) -> Allele:
        """Deletes the Gene at the index. If index is None, a Gene is
            deleted at a random index. Returns self for chaining
            operations.
        """
        index = index if index is not None else randint(0, len(self.genes)-1)
        del self.genes[index]
        return self

    def substitute(self, index: int = None, gene: Gene = None, *,
                   n_bases: int = None, max_base_size: int = 10,
                   base_factory: Callable[[None], int|float|str] = None,
                   factory_args: list = [], factory_kwargs: dict = {}) -> Allele:
        """Substitutes the Gene at the index with the given gene. If
            index is None, a random index will be used. If gene is None,
            adds a random gene using Gene.make, passing the kwargs.
            Returns self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.genes)-1)
        typert(index, int, "index")
        vert(index < len(self.genes), "index out of range")

        if gene is None:
            n_bases = n_bases or randint(0, max(len(g.bases) for g in self.genes))
            typert(n_bases, int, "n_bases")
            gene = Gene.make(
                n_bases, max_base_size=max_base_size, base_factory=base_factory,
                factory_args=factory_args,factory_kwargs=factory_kwargs
            )
        typert(gene, Gene, "gene")

        self.genes[index] = gene
        return self

    def recombine(self, other: Allele, indices: list[int] = None,
                  recombine_genes: bool = True,
                  match_genes: bool = True) -> Allele:
        """Recombines with the other Allele, swapping at the given
            indices. If indices is None, between 1 and ceil(log(len(self.genes)))
            random indices will be chosen. Recombines individual Genes
            if recombine_genes is True. Recombines only Genes with
            matching names if match_genes is True. Returns the new
            Allele.
        """
        typert(other, Allele, "other")
        vert(len(other.genes) > 0, "other must have genes")
        max_size = min(len(self.genes), len(other.genes))
        max_swaps = ceil(log(max_size)) or 1
        tert(indices is None or type(indices) is list,
             "indices must be list[int] or None")
        if type(indices) is list:
            tert(all(type(i) is int for i in indices),
                 "indices must be list[int] or None")
            vert(len(indices) <= max_size,
                 f"can have at most {max_size} indices")
        else:
            swaps = randint(0, max_swaps)
            indices = list(set([randint(0, max_size-1) for _ in range(swaps)]))
            indices.sort()

        name = self.name
        if self.name != other.name:
            name_size = min(len(self.name), len(other.name))
            name_swap = randint(1, name_size-1)
            name = self.name[:name_swap] + other.name[name_swap:]

        genes = [*self.genes]
        other_genes = [*other.genes]
        swapped = False
        for i in indices:
            genes[i:] = self.genes[i:] if swapped else other.genes[i:]
            other_genes[i:] = other.genes[i:] if swapped else self.genes[i:]
        genes = [g.copy() for g in genes]
        other_genes = [g.copy() for g in other_genes]

        if recombine_genes:
            for i in range(max_size):
                if genes[i].name == other_genes[i].name or not match_genes:
                    genes[i] = genes[i].recombine(other_genes[i])

        return Allele(name=name, genes=genes)

    @classmethod
    def make(cls, n_genes: int, n_bases: int, name: str = None, *,
             max_base_size: int = 10,
             base_factory: Callable[[None], int|float|str] = None,
             factory_args: list = [], factory_kwargs: dict = {}) -> Allele:
        """Makes and returns an Allele of randomized Genes."""
        genes = [
            Gene.make(
                n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
            for _ in range(n_genes)
        ]
        if name:
            return cls(name=name, genes=genes)
        return cls(genes=genes)

    def to_dict(self) -> dict:
        """Serialize the Allele to a dict."""
        return {
            self.name: [gene.to_dict() for gene in self.genes],
        }

    @classmethod
    def from_dict(cls, data: dict) -> Allele:
        """Deserialize an Allele from a dict."""
        for name, genes in data.items():
            unpacked = [Gene.from_dict(d) for d in genes]
            return cls(name=name, genes=unpacked)

    def __hash__(self) -> int:
        """Make Allele hashable."""
        return hash((self.name, tuple(self.genes)))


@dataclass
class Chromosome:
    name: str = field(default_factory=lambda: random_str(2))
    alleles: list[Allele] = field(default_factory=list)

    def copy(self) -> Chromosome:
        """Returns an exact copy of the Chromosome, copying the
            underlying Alleles as well.
        """
        return Chromosome(name=self.name, alleles=[a.copy() for a in self.alleles])

    def insert(self, index: int = None, allele: Allele = None, *,
               n_genes: int = None, n_bases: int = None, max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Chromosome:
        """Inserts the allele at the index. If index is None, the allele
            is inserted at a random index. If allele is None, adds a
            random allele using Allele.make, passing the kwargs. Returns
            self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.alleles)-1)
        typert(index, int, "index")
        vert(0 <= index <= len(self.alleles), "index out of range")

        if allele is None:
            if n_genes is None:
                n_genes = randint(0, max(len(a.genes) for a in self.alleles))
            if n_bases is None:
                n_bases = randint(0, max(len(g.bases) for a in self.alleles for g in a.genes))
            typert(n_genes, int, "n_genes")
            typert(n_bases, int, "n_bases")
            allele = Allele.make(
                n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
        typert(allele, Allele, "allele")

        self.alleles.insert(index, allele)
        return self

    def append(self, allele: Allele = None, *, n_genes: int = None,
               n_bases: int = None, max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Chromosome:
        """Adds a allele to the end of the chromosome. If allele is None,
            adds a random Allele using Allele.make and passing kwargs.
            Returns self for chaining operations.
        """
        return self.insert(
            len(self.alleles), allele, n_genes=n_genes, n_bases=n_bases,
            max_base_size=max_base_size, base_factory=base_factory,
            factory_args=factory_args, factory_kwargs=factory_kwargs
        )

    def duplicate(self, index: int = None) -> Chromosome:
        """Duplicates the Allele at the index. If index is None, the
            Allele at a random index is duplicated. Returns self for
            chaining operations.
        """
        index = index if index is not None else randint(0, len(self.alleles)-1)
        typert(index, int, "index")
        vert(0 <= index < len(self.alleles), "index out of range")
        return self.insert(index, self.alleles[index].copy())

    def delete(self, index: int = None) -> Chromosome:
        """Deletes the Allele at the index. If index is None, an Allele
            is deleted at a random index. Returns self for chaining
            operations.
        """
        index = index if index is not None else randint(0, len(self.alleles)-1)
        del self.alleles[index]
        return self

    def substitute(self, index: int = None, allele: Allele = None, *,
                   n_genes: int = None, n_bases: int = None,
                   max_base_size: int = 10,
                   base_factory: Callable[[None], int|float|str] = None,
                   factory_args: list = [], factory_kwargs: dict = {}) -> Chromosome:
        """Substitutes the Allele at the index with the given allele. If
            index is None, a random index will be used. If allele is
            None, adds a random allele using Allele.make, passing the
            kwargs. Returns self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.alleles)-1)
        typert(index, int, "index")
        vert(index < len(self.alleles), "index out of range")

        if allele is None:
            if n_genes is None:
                n_genes = randint(0, max(len(a.genes) for a in self.alleles))
            if n_bases is None:
                n_bases = randint(0, max(len(g.bases) for a in self.alleles for g in a.genes))
            typert(n_genes, int, "n_genes")
            typert(n_bases, int, "n_bases")
            allele = Allele.make(
                n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
        typert(allele, Allele, "allele")

        self.alleles[index] = allele
        return self

    def recombine(self, other: Chromosome, indices: list[int] = None,
                  recombine_alleles: bool = True, match_alleles: bool = True,
                  recombine_genes: bool = True, match_genes: bool = True
                  ) -> Chromosome:
        """Recombines with the other Chromosome, swapping at the given
            indices. If indices is None, between 1 and ceil(log(len(self.alleles)))
            random indices will be chosen. Recombines individual Alleles
            if recombine_allels is True. Recombines only Alleles with
            matching names if match_alleles is True. Recombines
            individual Genes if recombine_genes is True. Recombines only
            Genes with matching names if match_genes is True. Returns
            the new Chromosome.
        """
        typert(other, Chromosome, "other")
        vert(len(other.alleles) > 0, "other must have alleles")
        max_size = min(len(self.alleles), len(other.alleles))
        max_swaps = ceil(log(max_size)) or 1
        tert(indices is None or type(indices) is list,
             "indices must be list[int] or None")
        if type(indices) is list:
            tert(all(type(i) is int for i in indices),
                 "indices must be list[int] or None")
            vert(len(indices) <= max_size,
                 f"can have at most {max_size} indices")
        else:
            swaps = randint(0, max_swaps)
            indices = list(set([randint(0, max_size-1) for _ in range(swaps)]))
            indices.sort()

        name = self.name
        if self.name != other.name:
            name_size = min(len(self.name), len(other.name))
            name_swap = randint(1, name_size-1)
            name = self.name[:name_swap] + other.name[name_swap:]

        alleles = [*self.alleles]
        other_alleles = [*other.alleles]
        swapped = False
        for i in indices:
            alleles[i:] = self.alleles[i:] if swapped else other.alleles[i:]
            other_alleles[i:] = other.alleles[i:] if swapped else self.alleles[i:]
        alleles = [g.copy() for g in alleles]
        other_alleles = [g.copy() for g in other_alleles]

        if recombine_alleles:
            for i in range(max_size):
                if alleles[i].name == other_alleles[i].name or not match_alleles:
                    alleles[i] = alleles[i].recombine(
                        other_alleles[i],
                        recombine_genes=recombine_genes,
                        match_genes=match_genes
                    )

        return Chromosome(name=name, alleles=alleles)

    @classmethod
    def make(cls, n_alleles: int, n_genes: int, n_bases: int, name: str = None,
             *, max_base_size: int = 10,
             base_factory: Callable[[None], int|float|str] = None,
             factory_args: list = [], factory_kwargs: dict = {}) -> Chromosome:
        """Makes and returns a Chromosome of randomized Alleles."""
        alleles = [
            Allele.make(
                n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
            for _ in range(n_alleles)
        ]
        if name:
            return cls(name=name, alleles=alleles)
        return cls(alleles=alleles)

    def to_dict(self) -> dict:
        """Serialize the Chromosome to a dict."""
        return {
            self.name: [allele.to_dict() for allele in self.alleles],
        }

    @classmethod
    def from_dict(cls, data: dict) -> Chromosome:
        """Deserialize a Chromosome from a dict."""
        for name, alleles in data.items():
            unpacked = [Allele.from_dict(d) for d in alleles]
            return cls(name=name, alleles=unpacked)

    def __hash__(self) -> int:
        """Make Chromosome hashable."""
        return hash((self.name, tuple(self.alleles)))


@dataclass
class Genome:
    name: str = field(default_factory=lambda: random_str(6))
    chromosomes: list[Chromosome] = field(default_factory=list)

    def copy(self) -> Genome:
        """Returns an exact copy of the Genome, copying the
            underlying Chromosome as well.
        """
        return Genome(name=self.name, chromosomes=[a.copy() for a in self.chromosomes])

    def insert(self, index: int = None, chromosome: Chromosome = None, *,
               n_alleles: int = None, n_genes: int = None, n_bases: int = None,
               max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Chromosome:
        """Inserts the chromosome at the index. If index is None, the chromosome
            is inserted at a random index. If chromosome is None, adds a
            random chromosome using Chromosome.make, passing the kwargs. Returns
            self for chaining operations.
        """
        index = index if index is not None else randint(0, len(self.chromosomes)-1)
        typert(index, int, "index")
        vert(0 <= index <= len(self.chromosomes), "index out of range")

        if chromosome is None:
            if n_alleles is None:
                n_alleles = randint(0, max(len(c.alleles) for c in self.chromosomes))
            if n_genes is None:
                n_genes = randint(0, max(
                    len(a.genes)
                    for c in self.chromosomes
                    for a in c.alleles
                ))
            if n_bases is None:
                n_bases = randint(0, max(
                    len(g.bases)
                    for c in self.chromosomes
                    for a in c.alleles
                    for g in a.genes
                ))
            typert(n_alleles, int, "n_alleles")
            typert(n_genes, int, "n_genes")
            typert(n_bases, int, "n_bases")
            chromosome = Chromosome.make(
                n_alleles, n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
        typert(chromosome, Chromosome, "chromosome")

        self.chromosomes.insert(index, chromosome)
        return self

    def append(self, chromosome: Chromosome = None, *, n_alleles: int = None,
               n_genes: int = None, n_bases: int = None, max_base_size: int = 10,
               base_factory: Callable[[None], int|float|str] = None,
               factory_args: list = [], factory_kwargs: dict = {}) -> Genome:
        """Adds a chromosome to the end of the genome. If chromosome is
            None, adds a random Chromosome using Chromosome.make and
            passing kwargs. Returns self for chaining operations.
        """
        return self.insert(
            len(self.chromosomes), chromosome, n_alleles=n_alleles,
            n_genes=n_genes, n_bases=n_bases, max_base_size=max_base_size,
            base_factory=base_factory, factory_args=factory_args,
            factory_kwargs=factory_kwargs
        )

    def duplicate(self, index: int = None) -> Genome:
        """Duplicates the Chromosome at the index. If index is None, the
            Chromosome at a random index is duplicated. Returns self for
            chaining operations.
        """
        index = index if index is not None else randint(0, len(self.chromosomes)-1)
        typert(index, int, "index")
        vert(0 <= index < len(self.chromosomes), "index out of range")
        return self.insert(index, self.chromosomes[index].copy())

    def delete(self, index: int = None) -> Genome:
        """Deletes the Chromosome at the index. If index is None, a
            Chromosome is deleted at a random index. Returns self for
            chaining operations.
        """
        index = index if index is not None else randint(0, len(self.chromosomes)-1)
        del self.chromosomes[index]
        return self

    def substitute(self, index: int = None, chromosome: Chromosome = None, *,
                   n_alleles: int = None, n_genes: int = None,
                   n_bases: int = None, max_base_size: int = 10,
                   base_factory: Callable[[None], int|float|str] = None,
                   factory_args: list = [], factory_kwargs: dict = {}) -> Genome:
        """Substitutes the Chromosome at the index with the given
            chromosome. If index is None, a random index will be used.
            If chromosome is None, adds a random chromosome using
            Chromosome.make, passing the kwargs. Returns self for
            chaining operations.
        """
        index = index if index is not None else randint(0, len(self.chromosomes)-1)
        typert(index, int, "index")
        vert(index < len(self.chromosomes), "index out of range")

        if chromosome is None:
            if n_alleles is None:
                n_alleles = randint(0, max(len(c.alleles) for c in self.chromosomes))
            if n_genes is None:
                n_genes = randint(0, max(
                    len(a.genes)
                    for c in self.chromosomes
                    for a in c.alleles
                ))
            if n_bases is None:
                n_bases = randint(0, max(
                    len(g.bases)
                    for c in self.chromosomes
                    for a in c.alleles
                    for g in a.genes
                ))
            typert(n_alleles, int, "n_alleles")
            typert(n_genes, int, "n_genes")
            typert(n_bases, int, "n_bases")
            chromosome = Chromosome.make(
                n_alleles, n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
        typert(chromosome, Chromosome, "chromosome")

        self.chromosomes[index] = chromosome
        return self

    def recombine(self, other: Genome, indices: list[int] = None,
                  recombine_chromosomes: bool = True,
                  match_chromosomes: bool = True,
                  recombine_alleles: bool = True, match_alleles: bool = True,
                  recombine_genes: bool = True, match_genes: bool = True
                  ) -> Genome:
        """Recombines with the other Genome, swapping at the given
            indices. If indices is None, between 1 and
            ceil(log(len(self.chromosomes))) random indices will be
            chosen. Recombines individual Chromosomes
            if recombine_chromosomes is True. Recombines only
            Chromosomes with matching names if match_chromosomes is True.
            Recombines individual Alleles if recombine_alleles is True.
            Recombines only Alleles with matching names if match_alleles
            is True. Recombines individual Genes if recombine_genes is
            True. Recombines only Genes with matching names if
            match_genes is True. Returns the new Genome.
        """
        typert(other, Genome, "other")
        vert(len(other.chromosomes) > 0, "other must have chromosomes")
        max_size = min(len(self.chromosomes), len(other.chromosomes))
        max_swaps = ceil(log(max_size)) or 1
        tert(indices is None or type(indices) is list,
             "indices must be list[int] or None")
        if type(indices) is list:
            tert(all(type(i) is int for i in indices),
                 "indices must be list[int] or None")
            vert(len(indices) <= max_size,
                 f"can have at most {max_size} indices")
        else:
            swaps = randint(0, max_swaps)
            indices = list(set([randint(0, max_size-1) for _ in range(swaps)]))
            indices.sort()

        name = self.name
        if self.name != other.name:
            name_size = min(len(self.name), len(other.name))
            name_swap = randint(1, name_size-1)
            name = self.name[:name_swap] + other.name[name_swap:]

        chromosomes = [*self.chromosomes]
        other_chromosomes = [*other.chromosomes]
        swapped = False
        for i in indices:
            chromosomes[i:] = self.chromosomes[i:] if swapped else other.chromosomes[i:]
            other_chromosomes[i:] = other.chromosomes[i:] if swapped else self.chromosomes[i:]
        chromosomes = [g.copy() for g in chromosomes]
        other_chromosomes = [g.copy() for g in other_chromosomes]

        if recombine_chromosomes:
            for i in range(max_size):
                if chromosomes[i].name == other_chromosomes[i].name or not match_chromosomes:
                    chromosomes[i] = chromosomes[i].recombine(
                        other_chromosomes[i],
                        recombine_alleles=recombine_alleles,
                        recombine_genes=recombine_genes,
                        match_alleles=match_alleles,
                        match_genes=match_genes
                    )

        return Genome(name=name, chromosomes=chromosomes)

    @classmethod
    def make(cls, n_chromosomes: int, n_alleles: int, n_genes: int,
             n_bases: int, name: str = None, *, max_base_size: int = 10,
             base_factory: Callable[[None], int|float|str] = None,
             factory_args: list = [], factory_kwargs: dict = {}) -> Genome:
        """Makes and returns a Genome of randomized Chromosomes."""
        chromosomes = [
            Chromosome.make(
                n_alleles, n_genes, n_bases, max_base_size=max_base_size,
                base_factory=base_factory, factory_args=factory_args,
                factory_kwargs=factory_kwargs
            )
            for _ in range(n_chromosomes)
        ]
        if name:
            return cls(name=name, chromosomes=chromosomes)
        return cls(chromosomes=chromosomes)

    def to_dict(self) -> dict:
        """Serialize the Genome to a dict."""
        return {
            self.name: [chromosome.to_dict() for chromosome in self.chromosomes],
        }

    @classmethod
    def from_dict(cls, data: dict) -> Genome:
        """Deserialize a Genome from a dict."""
        for name, chromosomes in data.items():
            unpacked = [Chromosome.from_dict(c) for c in chromosomes]
            return cls(name=name, chromosomes=unpacked)

    def __hash__(self) -> int:
        """Make Genome hashable."""
        return hash((self.name, tuple(self.chromosomes)))



from context import genes, optimization
from random import randint, random
import unittest


class TestOptimizationForGenes(unittest.TestCase):
    def population(self) -> list[genes.Gene]:
        return [
            genes.Gene("test", [1.0, 2.0, 3.0, 4.0]),
            genes.Gene("test", [2.1, 1.1, 4.1, 3.1]),
            genes.Gene("test", [0.2, 3.2, 2.2, 1.2]),
            genes.Gene("test", [3.3, 4.3, 1.3, 2.3]),
        ]

    def setUp(self) -> None:
        self.parents = self.population()
        return super().setUp()

    def test_child_from_parents_returns_recombination_of_two_random_parents(self):
        child = optimization.child_from_parents(self.parents)

        parent_bases = {
            i: self.parents[i].bases
            for i in range(len(self.parents))
        }
        parents = set()
        for base in child.bases:
            for k, v in parent_bases.items():
                if base in v:
                    parents.add(k)
        assert len(parents) <= 2

    def test_optimize_gene_exits_after_max_iterations_or_achieving_fitness_target(self):
        target = 123.456

        def measure_fitness(gene: genes.Gene) -> float:
            return 1 / abs(sum(gene.bases) - target)

        def mutate_gene(gene: genes.Gene) -> genes.Gene:
            for i in range(len(gene.bases)):
                val = random()
                if val <= 0.1:
                    gene.bases[i] *= random()
                elif val <= 0.2:
                    gene.bases[i] /= random()
                elif val <= 0.6:
                    gene.bases[i] += random()
                else:
                    gene.bases[i] -= random()
            return gene

        count, population = optimization.optimize_gene(
            measure_fitness=measure_fitness,
            mutate_gene=mutate_gene,
            initial_population=self.population(),
            max_iterations=100,
            population_size=100
        )

        assert type(count) is int and count <= 100
        assert type(population) is list
        assert all(type(p) is genes.Gene for p in population)
        assert len(population) == 100
        best = (sum(population[0].bases), population[0])

        assert count < 100 or (best[0] - target)/target < 0.01


class TestOptimizationForAlleles(unittest.TestCase):
    def population(self) -> list[genes.Allele]:
        return [
            genes.Allele("test", [
                genes.Gene("gn1", [1.0, 2.0, 3.0, 4.0]),
                genes.Gene("gn2", [2.1, 1.1, 4.1, 3.1]),
            ]),
            genes.Allele("test", [
                genes.Gene("gn1", [1.2, 2.2, 3.2, 4.2]),
                genes.Gene("gn2", [2.3, 1.3, 4.3, 3.3]),
            ]),
            genes.Allele("test", [
                genes.Gene("gn1", [1.4, 2.4, 3.4, 4.4]),
                genes.Gene("gn2", [2.5, 1.5, 4.5, 3.5]),
            ]),
            genes.Allele("test", [
                genes.Gene("gn1", [1.6, 2.6, 3.6, 4.6]),
                genes.Gene("gn2", [2.7, 1.7, 4.7, 3.7]),
            ]),
        ]

    def setUp(self) -> None:
        self.parents = self.population()
        return super().setUp()

    def test_child_from_parents_returns_recombination_of_two_random_parents(self):
        child = optimization.child_from_parents(self.parents)

        parent_genes = {
            i: self.parents[i].genes
            for i in range(len(self.parents))
        }
        parent_bases = {
            i: [b for g in pg for b in g.bases]
            for i, pg in parent_genes.items()
        }
        parents = set()
        for gene in child.genes:
            for base in gene.bases:
                for k, v in parent_bases.items():
                    if base in v:
                        parents.add(k)
        assert len(parents) <= 2

    def test_optimize_allele_exits_after_max_iterations_or_achieving_fitness_target(self):
        target = 123.456

        def measure_fitness(allele: genes.Allele) -> float:
            sums = [sum(g.bases) for g in allele.genes]
            return 1 / (1 + abs(sum(sums) - target))

        def mutate_gene(gene: genes.Gene) -> genes.Gene:
            for i in range(len(gene.bases)):
                val = random()
                if val <= 0.1:
                    gene.bases[i] *= random()
                elif val <= 0.2:
                    gene.bases[i] /= random()
                elif val <= 0.6:
                    gene.bases[i] += random()
                else:
                    gene.bases[i] -= random()
            return gene

        def mutate_allele(allele: genes.Allele) -> genes.Allele:
            allele.genes = [mutate_gene(g) for g in allele.genes]
            return allele

        count, population = optimization.optimize_allele(
            measure_fitness=measure_fitness,
            mutate_allele=mutate_allele,
            initial_population=self.population(),
            max_iterations=100,
            population_size=100
        )

        assert type(count) is int and count <= 100
        assert type(population) is list
        assert all(type(p) is genes.Allele for p in population)
        assert len(population) == 100
        sums = [sum(g.bases) for g in population[0].genes]
        best = (sum(sums), population[0])

        assert count < 100 or (best[0] - target)/target < 0.01


class TestOptimizationForChromosomes(unittest.TestCase):
    def population(self) -> list[genes.Chromosome]:
        return [
            genes.Chromosome("test", [
                genes.Allele("al1", [
                    genes.Gene("gn1", [1.0, 2.0, 3.0, 4.0]),
                    genes.Gene("gn2", [2.1, 1.1, 4.1, 3.1]),
                ]),
                genes.Allele("al2", [
                    genes.Gene("gn1", [11.0, 12.0, 13.0, 14.0]),
                    genes.Gene("gn2", [12.1, 11.1, 14.1, 13.1]),
                ]),
            ]),
            genes.Chromosome("test", [
                genes.Allele("al1", [
                    genes.Gene("gn1", [1.2, 2.2, 3.2, 4.2]),
                    genes.Gene("gn2", [2.3, 1.3, 4.3, 3.3]),
                ]),
                genes.Allele("al2", [
                    genes.Gene("gn1", [11.2, 12.2, 13.2, 14.2]),
                    genes.Gene("gn2", [12.3, 11.3, 14.3, 13.3]),
                ]),
            ]),
            genes.Chromosome("test", [
                genes.Allele("al1", [
                    genes.Gene("gn1", [1.4, 2.4, 3.4, 4.4]),
                    genes.Gene("gn2", [2.5, 1.5, 4.5, 3.5]),
                ]),
                genes.Allele("al2", [
                    genes.Gene("gn1", [11.4, 12.4, 13.4, 14.4]),
                    genes.Gene("gn2", [12.5, 11.5, 14.5, 13.5]),
                ]),
            ]),
            genes.Chromosome("test", [
                genes.Allele("al1", [
                    genes.Gene("gn1", [1.6, 2.6, 3.6, 4.6]),
                    genes.Gene("gn2", [2.7, 1.7, 4.7, 3.7]),
                ]),
                genes.Allele("al2", [
                    genes.Gene("gn1", [11.6, 12.6, 13.6, 14.6]),
                    genes.Gene("gn2", [12.7, 11.7, 14.7, 13.7]),
                ]),
            ]),
        ]

    def setUp(self) -> None:
        self.parents = self.population()
        return super().setUp()

    def test_child_from_parents_returns_recombination_of_two_random_parents(self):
        child = optimization.child_from_parents(self.parents)

        parent_alleles = {
            i: self.parents[i].alleles
            for i in range(len(self.parents))
        }
        parent_genes = {
            i: [g for a in pa for g in a.genes]
            for i, pa in parent_alleles.items()
        }
        parent_bases = {
            i: [b for g in pg for b in g.bases]
            for i, pg in parent_genes.items()
        }
        parents = set()
        for allele in child.alleles:
            for gene in allele.genes:
                for base in gene.bases:
                    for k, v in parent_bases.items():
                        if base in v:
                            parents.add(k)
        assert len(parents) <= 2

    def test_optimize_chromosome_exits_after_max_iterations_or_achieving_fitness_target(self):
        target = 123.456

        def measure_fitness(chromosome: genes.Chromosome) -> float:
            sums = [sum(g.bases) for a in chromosome.alleles for g in a.genes]
            return 1 / (1 + abs(sum(sums) - target))

        def mutate_gene(gene: genes.Gene) -> genes.Gene:
            for i in range(len(gene.bases)):
                val = random()
                if val <= 0.1:
                    gene.bases[i] *= random()
                elif val <= 0.2:
                    gene.bases[i] /= random()
                elif val <= 0.6:
                    gene.bases[i] += random()
                else:
                    gene.bases[i] -= random()
            return gene

        def mutate_allele(allele: genes.Allele) -> genes.Allele:
            allele.genes = [mutate_gene(g) for g in allele.genes]
            return allele

        def mutate_chromosome(chromosome: genes.Chromosome) -> genes.Chromosome:
            chromosome.alleles = [mutate_allele(a) for a in chromosome.alleles]
            return chromosome

        count, population = optimization.optimize_chromosome(
            measure_fitness=measure_fitness,
            mutate_chromosome=mutate_chromosome,
            initial_population=self.population(),
            max_iterations=100,
            population_size=100
        )

        assert type(count) is int and count <= 100
        assert type(population) is list
        assert all(type(p) is genes.Chromosome for p in population)
        assert len(population) == 100
        sums = [sum(g.bases) for a in population[0].alleles for g in a.genes]
        best = (sum(sums), population[0])

        assert count < 100 or (best[0] - target)/target < 0.01


class TestOptimizationForGenomes(unittest.TestCase):
    def population(self) -> list[genes.Genome]:
        return [
            genes.Genome("test", [
                genes.Chromosome("chr1", [
                    genes.Allele("al1", [
                        genes.Gene("gn1", [1.0, 2.0, 3.0, 4.0]),
                        genes.Gene("gn2", [2.1, 1.1, 4.1, 3.1]),
                    ]),
                    genes.Allele("al2", [
                        genes.Gene("gn1", [11.0, 12.0, 13.0, 14.0]),
                        genes.Gene("gn2", [12.1, 11.1, 14.1, 13.1]),
                    ]),
                ])
            ]),
            genes.Genome("test", [
                genes.Chromosome("chr1", [
                    genes.Allele("al1", [
                        genes.Gene("gn1", [1.2, 2.2, 3.2, 4.2]),
                        genes.Gene("gn2", [2.3, 1.3, 4.3, 3.3]),
                    ]),
                    genes.Allele("al2", [
                        genes.Gene("gn1", [11.2, 12.2, 13.2, 14.2]),
                        genes.Gene("gn2", [12.3, 11.3, 14.3, 13.3]),
                    ]),
                ])
            ]),
            genes.Genome("test", [
                genes.Chromosome("chr1", [
                    genes.Allele("al1", [
                        genes.Gene("gn1", [1.4, 2.4, 3.4, 4.4]),
                        genes.Gene("gn2", [2.5, 1.5, 4.5, 3.5]),
                    ]),
                    genes.Allele("al2", [
                        genes.Gene("gn1", [11.4, 12.4, 13.4, 14.4]),
                        genes.Gene("gn2", [12.5, 11.5, 14.5, 13.5]),
                    ]),
                ])
            ]),
            genes.Genome("test", [
                genes.Chromosome("chr1", [
                    genes.Allele("al1", [
                        genes.Gene("gn1", [1.6, 2.6, 3.6, 4.6]),
                        genes.Gene("gn2", [2.7, 1.7, 4.7, 3.7]),
                    ]),
                    genes.Allele("al2", [
                        genes.Gene("gn1", [11.6, 12.6, 13.6, 14.6]),
                        genes.Gene("gn2", [12.7, 11.7, 14.7, 13.7]),
                    ]),
                ])
            ]),
        ]

    def setUp(self) -> None:
        self.parents = self.population()
        return super().setUp()

    def test_child_from_parents_returns_recombination_of_two_random_parents(self):
        child = optimization.child_from_parents(self.parents)

        parent_chromosomes = {
            i: self.parents[i].chromosomes
            for i in range(len(self.parents))
        }
        parent_alleles = {
            i: [a for c in pc for a in c.alleles]
            for i, pc in parent_chromosomes.items()
        }
        parent_genes = {
            i: [g for a in pa for g in a.genes]
            for i, pa in parent_alleles.items()
        }
        parent_bases = {
            i: [b for g in pg for b in g.bases]
            for i, pg in parent_genes.items()
        }
        parents = set()
        for chromosome in child.chromosomes:
            for allele in chromosome.alleles:
                for gene in allele.genes:
                    for base in gene.bases:
                        for k, v in parent_bases.items():
                            if base in v:
                                parents.add(k)
        assert len(parents) <= 2

    def test_optimize_genome_exits_after_max_iterations_or_achieving_fitness_target(self):
        target = 123.456

        def measure_fitness(genome: genes.Genome) -> float:
            sums = [
                sum(g.bases)
                for c in genome.chromosomes
                for a in c.alleles
                for g in a.genes
            ]
            return 1 / (1 + abs(sum(sums) - target))

        def mutate_gene(gene: genes.Gene) -> genes.Gene:
            for i in range(len(gene.bases)):
                val = random()
                if val <= 0.1:
                    gene.bases[i] *= random()
                elif val <= 0.2:
                    gene.bases[i] /= random()
                elif val <= 0.6:
                    gene.bases[i] += random()
                else:
                    gene.bases[i] -= random()
            return gene

        def mutate_allele(allele: genes.Allele) -> genes.Allele:
            allele.genes = [mutate_gene(g) for g in allele.genes]
            return allele

        def mutate_chromosome(chromosome: genes.Chromosome) -> genes.Chromosome:
            chromosome.alleles = [mutate_allele(a) for a in chromosome.alleles]
            return chromosome

        def mutate_genome(genome: genes.Genome) -> genes.Genome:
            genome.chromosomes = [mutate_chromosome(c) for c in genome.chromosomes]
            return genome

        count, population = optimization.optimize_genome(
            measure_fitness=measure_fitness,
            mutate_genome=mutate_genome,
            initial_population=self.population(),
            max_iterations=100,
            population_size=100
        )

        assert type(count) is int and count <= 100
        assert type(population) is list
        assert all(type(p) is genes.Genome for p in population)
        assert len(population) == 100
        sums = [
            sum(g.bases)
            for c in population[0].chromosomes
            for a in c.alleles
            for g in a.genes
        ]
        best = (sum(sums), population[0])

        assert count < 100 or (best[0] - target)/target < 0.01


class TestOptimizationHookForGenes(unittest.TestCase):
    def setUp(self) -> None:
        self.original_hook = optimization._hook
        return super().setUp()

    def tearDown(self) -> None:
        optimization._hook = self.original_hook
        return super().tearDown()

    def test_set_hook_sets__hook(self):
        def hook (count: int, population: list[tuple[int, genes.Gene]]) -> None:
            print((count, population))

        assert optimization._hook is None
        optimization.set_hook(hook)
        assert optimization._hook is hook

    def test_unset_hook_unsets__hook(self):
        def hook (count: int, population: list[tuple[int, genes.Gene]]) -> None:
            print((count, population))

        assert optimization._hook is None
        optimization.set_hook(hook)
        assert optimization._hook is hook
        optimization.unset_hook()
        assert optimization._hook is None

    def test_hook_is_called_on_each_generation(self):
        logs = []
        def hook (count: int, population: list[tuple[int, genes.Gene]]) -> None:
            logs.append((count, population))

        target = 123
        def measure_fitness(gene: genes.Gene):
            return 1 / (1 + abs(sum(gene.bases) - target))

        def mutate_gene(gene: genes.Gene) -> genes.Gene:
            for i in range(len(gene.bases)):
                val = random()
                if val <= 0.5:
                    gene.bases[i] += randint(0, 10)
                else:
                    gene.bases[i] -= randint(0, 10)
            return gene

        optimization.set_hook(hook)
        count, _ = optimization.optimize_gene(
            measure_fitness, mutate_gene, max_iterations=100
        )

        assert len(logs) == count
        assert all([type(l) is tuple for l in logs])
        assert all([type(l[0]) is int for l in logs])
        assert all([type(l[1]) is list for l in logs])
        assert all([type(fs[0]) is float for l in logs for fs in l[1]])
        assert all([type(fs[1]) is genes.Gene for l in logs for fs in l[1]])


if __name__ == '__main__':
    unittest.main()



from .genes import (
    Gene,
    Allele,
    Chromosome,
    Genome
)
from .optimization import (
    optimize_gene,
    optimize_allele,
    optimize_chromosome,
    optimize_genome,
    set_hook,
    unset_hook,
)


from .errors import tert
from .genes import Gene, Allele, Chromosome, Genome
from random import choices
from typing import Any, Callable


T = Genome | Chromosome | Allele | Gene

_hook = None


def set_hook(func: Callable[[int, list[tuple[int|float, T]]], None]) -> None:
    """Set a hook that will be called after every generation and passed
        the current generation count and the population with fitness
        scores.
    """
    tert(callable(func), "func must be Callable")
    global _hook
    _hook = func


def unset_hook() -> None:
    """Unsets the hook."""
    global _hook
    _hook = None


def _optimize(
        opt_type: type,
        measure_fitness: Callable[[T, int|float], int|float],
        mutate_func: Callable[[T], T],
        initial_population: list[Chromosome] = None, population_size: int = 100,
        genome_size: int = 2, chromosome_size: int = 2, allele_size: int = 3,
        gene_size: int = 10, fitness_target: int|float = 1.0,
        max_iterations: int = 1000,
        base_factory: Callable[[Any], int|float|str] = None,
        factory_args: list[Any] = None,
        factory_kwargs: dict[str, Any] = None,
        T_name: str = None, parents_per_generation: int = 10,
    ) -> tuple[int, list[T]]:
    """Internal function for abstracting optimization code."""
    tert(opt_type in (Genome, Chromosome, Allele, Gene),
         "opt_type must be one of Genome, Chromosome, Allele, or Gene")
    tert(initial_population is None or all(type(p) is opt_type for p in initial_population),
         f"initial_population must be None or list[{opt_type.__name__}]")
    population = initial_population
    if population is None:
        if opt_type is Genome:
            population = [
                Genome.make(
                    genome_size, chromosome_size, allele_size, gene_size, T_name,
                    base_factory=base_factory, factory_args=factory_args,
                    factory_kwargs=factory_kwargs
                )
                for _ in range(population_size)
            ]
        elif opt_type is Chromosome:
            population = [
                Chromosome.make(
                    chromosome_size, allele_size, gene_size, T_name,
                    base_factory=base_factory, factory_args=factory_args,
                    factory_kwargs=factory_kwargs
                )
                for _ in range(population_size)
            ]
        elif opt_type is Allele:
            population = [
                Allele.make(
                    allele_size, gene_size, T_name, base_factory=base_factory,
                    factory_args=factory_args, factory_kwargs=factory_kwargs
                )
                for _ in range(population_size)
            ]
        else:
            population = [
                Gene.make(
                    gene_size, T_name, base_factory=base_factory,
                    factory_args=factory_args, factory_kwargs=factory_kwargs
                )
                for _ in range(population_size)
            ]

    count = 0
    fitness_scores: list[tuple[int|float, T]] = [
        (measure_fitness(g), g)
        for g in population
    ]
    fitness_scores.sort(key=lambda fs: fs[0])
    fitness_scores.reverse()
    best_fitness = fitness_scores[0][0]

    while count < max_iterations and best_fitness < fitness_target:
        # breed parents at random proportional to their order by score
        parents = [fs[1] for fs in fitness_scores[:parents_per_generation]]
        children = [
            child_from_parents(parents)
            for _ in range(population_size-len(parents))
        ]

        children = [mutate_func(child) for child in children]
        population = [*children, *parents]
        fitness_scores: list[tuple[int|float, T]] = [
            (measure_fitness(g), g)
            for g in population
        ]
        fitness_scores.sort(key=lambda fs: fs[0])
        fitness_scores.reverse()
        best_fitness = fitness_scores[0][0]
        count += 1

        if _hook:
            _hook(count, fitness_scores)

    population = [fs[1] for fs in fitness_scores]
    return (count, population)


def optimize_gene(
        measure_fitness: Callable[[Gene, int|float], int|float],
        mutate_gene: Callable[[Gene], Gene],
        initial_population: list[Gene] = None, population_size: int = 100,
        gene_size: int = 10, fitness_target: int|float = 1.0,
        max_iterations: int = 1000,
        base_factory: Callable[[Any], int|float|str] = None,
        factory_args: list[Any] = None,
        factory_kwargs: dict[str, Any] = None, gene_name: str = None,
        parents_per_generation: int = 10) -> tuple[int, list[Gene]]:
    """Optimize a Gene given a measure_fitness function, a mutate_gene
        function, a population_size int, a gene_size int, a
        fitness_target float, and a max_iterations int. Supply
        base_factory to produce Gene bases other than random ints
        between 0 and 10, with optional factory_args and factory_kwargs
        which will be passed to each call of base_factory. Supply
        gene_name to assign the name to each Gene in the population.
        Returns the number of iterations and the final population.
    """
    return _optimize(
        Gene, measure_fitness, mutate_gene, initial_population,
        population_size, None, None, None, gene_size, fitness_target,
        max_iterations, base_factory, factory_args, factory_kwargs, gene_name,
        parents_per_generation
    )


def optimize_allele(
        measure_fitness: Callable[[Allele, int|float], int|float],
        mutate_allele: Callable[[Allele], Allele],
        initial_population: list[Allele] = None, population_size: int = 100,
        allele_size: int = 2, gene_size: int = 10,
        fitness_target: int|float = 1.0, max_iterations: int = 1000,
        base_factory: Callable[[Any], int|float|str] = None,
        factory_args: list[Any] = None, factory_kwargs: dict[str, Any] = None,
        allele_name: str = None, parents_per_generation: int = 10
        ) -> tuple[int, list[Allele]]:
    """Optimize an Allele given a measure_fitness function, a
        mutate_allele function, a population_size int, a fitness_target
        float, and a max_iterations int. Supply base_factory to produce
        Gene bases other than random ints between 0 and 10,
        with optional factory_args and factory_kwargs which will be
        passed to each call of base_factory. Supply allele_name to
        assign the name to each generated Allele in the population.
        Supply an allele_size int and a gene_size int to customize
        generation of a random initial population, or supply
        initial_population list[Allele] to specify the initial
        population. Returns the number of iterations and the final
        population.
    """
    return _optimize(
        Allele, measure_fitness, mutate_allele, initial_population,
        population_size, None, None, allele_size, gene_size,
        fitness_target, max_iterations, base_factory, factory_args,
        factory_kwargs, allele_name, parents_per_generation
    )


def optimize_chromosome(
        measure_fitness: Callable[[Chromosome, int|float], int|float],
        mutate_chromosome: Callable[[Chromosome], Chromosome],
        initial_population: list[Chromosome] = None, population_size: int = 100,
        chromosome_size: int = 2, allele_size: int = 3, gene_size: int = 10,
        fitness_target: int|float = 1.0, max_iterations: int = 1000,
        base_factory: Callable[[Any], int|float|str] = None,
        factory_args: list[Any] = None,
        factory_kwargs: dict[str, Any] = None,
        chromosome_name: str = None, parents_per_generation: int = 10,
    ) -> tuple[int, list[Chromosome]]:
    """Optimize an Chromosome given a measure_fitness function, a
        mutate_chromosome function, a population_size int, a fitness_target
        float, and a max_iterations int. Supply base_factory to produce
        Gene bases other than random ints between 0 and 10,
        with optional factory_args and factory_kwargs which will be
        passed to each call of base_factory. Supply chromosome_name to
        assign the name to each generated Chromosome in the population.
        Supply an allele_size int and a gene_size int to customize
        generation of a random initial population, or supply
        initial_population list[Chromosome] to specify the initial
        population. Returns the number of iterations and the final
        population.
    """
    return _optimize(
        Chromosome, measure_fitness, mutate_chromosome, initial_population,
        population_size, None, chromosome_size, allele_size, gene_size,
        fitness_target, max_iterations, base_factory, factory_args,
        factory_kwargs, chromosome_name, parents_per_generation
    )


def optimize_genome(
        measure_fitness: Callable[[Genome, int|float], int|float],
        mutate_genome: Callable[[Genome], Genome],
        initial_population: list[Genome] = None, population_size: int = 100,
        genome_size: int = 1, chromosome_size: int = 2, allele_size: int = 3,
        gene_size: int = 10, fitness_target: int|float = 1.0,
        max_iterations: int = 1000,
        base_factory: Callable[[Any], int|float|str] = None,
        factory_args: list[Any] = None,
        factory_kwargs: dict[str, Any] = None,
        genome_name: str = None, parents_per_generation: int = 10,
    ) -> tuple[int, list[Genome]]:
    """Optimize a Genome given a measure_fitness function, a
        mutate_chromosome function, a population_size int, a fitness_target
        float, and a max_iterations int. Supply base_factory to produce
        Gene bases other than random ints between 0 and 10,
        with optional factory_args and factory_kwargs which will be
        passed to each call of base_factory. Supply genome_name to
        assign the name to each generated Genome in the population.
        Supply an allele_size int and a gene_size int to customize
        generation of a random initial population, or supply
        initial_population list[Genome] to specify the initial
        population. Returns the number of iterations and the final
        population.
    """
    return _optimize(
        Genome, measure_fitness, mutate_genome, initial_population,
        population_size, genome_size, chromosome_size, allele_size, gene_size,
        fitness_target, max_iterations, base_factory, factory_args,
        factory_kwargs, genome_name, parents_per_generation
    )


def child_from_parents(parents: list[Genome]) -> Genome:
    """Select two parents at random semi-proportional to their order in
        the list. Recombine the two chosen parent Genes, and return the
        result.
    """
    weights = [len(parents[i:])/len(parents) for i in range(len(parents))]
    dad, mom = choices(parents, weights, k=2)
    return dad.recombine(mom)



from context import genes
from random import random
import unittest


class TestGene(unittest.TestCase):
    def setUp(self) -> None:
        self.gene = genes.Gene("test", [1,2,3])
        return super().setUp()

    def test_Gene_is_hashable(self):
        hash(self.gene)

    def test_copy_returns_identical_Gene(self):
        copy = self.gene.copy()
        assert type(copy) is genes.Gene
        assert copy.name == self.gene.name
        assert copy.bases == self.gene.bases
        assert copy.bases is not self.gene.bases

    def test_insert_inserts_base_at_specified_or_random_index(self):
        self.gene.insert(0, 5)
        assert self.gene.bases[0] == 5
        self.gene.insert(3, 7)
        assert self.gene.bases[3] == 7

        indices = set()
        for _ in range(10):
            gene = genes.Gene("test", [1,2,3])
            gene.insert(base=4)
            indices.add(gene.bases.index(4))
        assert len(indices) > 1

    def test_append_adds_base_to_end_of_bases(self):
        self.gene.append(9)
        assert self.gene.bases[-1] == 9
        self.gene.append(10)
        assert self.gene.bases[-1] == 10

    def test_insert_sequence_at_specified_or_random_index(self):
        self.gene.insert_sequence(1, [8,9])
        assert self.gene.bases[1:][:2] == [8,9]

        indices = set()
        for _ in range(10):
            gene = genes.Gene('test', [1,2,3])
            gene.insert_sequence(sequence=[4,5])
            assert 4 in gene.bases
            assert 5 in gene.bases
            assert gene.bases.index(4) + 1 == gene.bases.index(5)
            indices.add(gene.bases.index(4))
        assert len(indices) > 1

    def test_delete_removes_base_at_specified_or_random_index(self):
        self.gene.delete(1)
        assert self.gene.bases == [1,3]

        results = set()
        for _ in range(10):
            gene = genes.Gene("test", [1,2,3])
            gene.delete()
            results.add(tuple(gene.bases))
        assert len(results) > 1

    def test_delete_sequence_removes_bases_at_specified_or_random_index_with_specified_or_random_size(self):
        gene = genes.Gene("big", list(range(10)))
        gene.delete_sequence(3, 5)
        assert gene.bases == [0, 1, 2, 8, 9]

        results = set()
        for _ in range(10):
            gene = genes.Gene("big", list(range(10)))
            gene.delete_sequence()
            results.add(tuple(gene.bases))
        assert len(results) > 1
        assert len([len(r) for r in results]) > 1

    def test_substitute_replaces_base_with_specified_or_random_base_at_specified_or_random_index(self):
        self.gene.substitute(1, 5)
        assert self.gene.bases[1] == 5

        results = set()
        for _ in range(10):
            gene = genes.Gene("big", list(range(10)))
            gene.substitute()
            results.add(tuple(gene.bases))
        assert len(results) > 1
        assert all(len(r) == 10 for r in results)

    def test_recombine_recombines_genes_at_specified_or_random_indices(self):
        gene1 = genes.Gene("base", list(range(5)))
        gene2 = genes.Gene("diff", list(range(5, 10)))
        gene3 = gene1.recombine(gene2, [1, 3])
        assert gene3.bases == [0, 6, 7, 3, 4]
        assert gene3.name != gene1.name
        assert gene3.name[0] == gene1.name[0]
        assert gene3.name != gene2.name
        assert gene3.name[-1] == gene2.name[-1]

        results = set()
        names = set()
        for _ in range(10):
            gene1 = genes.Gene("base", list(range(10)))
            gene2 = genes.Gene("diff", list(range(10, 20)))
            gene3 = gene1.recombine(gene2)
            results.add(tuple(gene3.bases))
            names.add(gene3.name)
        assert len(results) > 1
        assert len(names) > 1
        for g in results:
            for base in g:
                assert base in list(range(20))

    def test_make_returns_random_Gene(self):
        gene = genes.Gene.make(5)
        assert type(gene) is genes.Gene
        assert len(gene.bases) == 5

        results = set()
        for _ in range(10):
            results.add(genes.Gene.make(10))
        assert len(results) > 8
        assert len(set([g.name for g in results])) >= 9
        assert len(set([tuple(g.bases) for g in results])) >= 9

    def test_make_with_base_factory_sets_bases_with_correct_type(self):
        float_factory = lambda: random()
        gene = genes.Gene.make(5, base_factory=float_factory)
        assert all(type(b) is float for b in gene.bases)
        assert len(set(gene.bases)) >= 4

        float_factory = lambda mult: random()*mult
        gene = genes.Gene.make(5, base_factory=float_factory, factory_args=[-2])
        assert all(-2 <= b <= 0 for b in gene.bases)

    def test_to_dict_and_from_dict_e2e(self):
        encoded = self.gene.to_dict()
        assert type(encoded) is dict
        decoded = genes.Gene.from_dict(encoded)
        assert type(decoded) is genes.Gene
        assert self.gene == decoded
        assert self.gene is not decoded

        float_factory = lambda: random()
        gene = genes.Gene.make(5, base_factory=float_factory)
        encoded = gene.to_dict()
        assert type(encoded) is dict
        decoded = genes.Gene.from_dict(encoded)
        assert type(decoded) is genes.Gene
        assert decoded == gene
        assert decoded is not gene

        str_factory = lambda: genes.random_str(5)
        gene = genes.Gene.make(5, base_factory=str_factory)
        encoded = gene.to_dict()
        assert type(encoded) is dict
        decoded = genes.Gene.from_dict(encoded)
        assert type(decoded) is genes.Gene
        assert decoded == gene
        assert decoded is not gene


class TestAllele(unittest.TestCase):
    def first_allele(self) -> genes.Allele:
        return genes.Allele("test", [
            genes.Gene("g1", [1, 2, 3]),
            genes.Gene("g2", [4, 5, 6]),
            genes.Gene("g3", [7, 8, 9]),
            genes.Gene("g4", [-7, -8, -9]),
        ])

    def alt_allele(self) -> genes.Allele:
        return genes.Allele("test", [
            genes.Gene("g1", [10, 11, 12]),
            genes.Gene("g2", [13, 14, 15]),
            genes.Gene("g3", [16, 17, 18]),
            genes.Gene("g4", [-16, -17, -18]),
        ])

    def second_allele(self) -> genes.Allele:
        return genes.Allele("diff", [
            genes.Gene("g4", [10, 11, 12]),
            genes.Gene("g5", [13, 14, 15]),
            genes.Gene("g6", [16, 17, 18]),
            genes.Gene("g6", [116, 117, 118]),
        ])

    def first_gene(self) -> genes.Gene:
        return genes.Gene("test", [1,2,3])

    def second_gene(self) -> genes.Gene:
        return genes.Gene("diff", [4,5,6])

    def setUp(self) -> None:
        self.allele = self.first_allele()
        self.alternate = self.alt_allele()
        self.other = self.second_allele()
        return super().setUp()

    def test_Allele_is_hashable(self):
        hash(self.allele)

    def test_copy_returns_identical_Allele_containing_identical_genes(self):
        copy = self.allele.copy()
        assert type(copy) is genes.Allele
        assert copy is not self.allele
        assert copy == self.allele

        for a, b in zip(copy.genes, self.allele.genes):
            assert a == b
            assert a.name == b.name
            assert a.bases == b.bases
            assert a is not b

    def test_insert_inserts_specified_or_random_gene_at_specified_or_random_index(self):
        g4 = genes.Gene("g4", [10, 11, 12])
        g5 = genes.Gene("g5", [13, 14, 15])
        self.allele.insert(0, g4)
        assert self.allele.genes[0] == g4
        self.allele.insert(2, g5)
        assert self.allele.genes[2] == g5

        indices = set()
        ref_gene = self.first_gene()
        for _ in range(10):
            allele = self.first_allele()
            allele.insert(gene=ref_gene)
            indices.add(allele.genes.index(ref_gene))
        assert len(indices) > 1

        gs = set()
        for _ in range(10):
            allele = self.first_allele()
            allele.insert(0)
            gs.add(allele.genes[0])
        assert len(gs) >= 4

    def test_append_adds_Gene_to_end_of_genes(self):
        gg1 = genes.Gene("gg1", [1,2,3])
        gg2 = genes.Gene("gg2", [1,2,3])
        self.allele.append(gg1)
        assert self.allele.genes[-1] == gg1
        self.allele.append(gg2)
        assert self.allele.genes[-1] == gg2

    def test_duplicate_duplicates_Gene_at_specified_or_random_index(self):
        self.allele.duplicate(0)
        assert self.allele.genes[0] == self.allele.genes[1]
        assert self.allele.genes[0] is not self.allele.genes[1]

        indices = set()
        for _ in range(10):
            allele = self.first_allele()
            allele.duplicate()
            for i in range(len(allele.genes)):
                if allele.genes[i] == allele.genes[i+1]:
                    indices.add(i)
                    break
        assert len(indices) >= 2

    def test_delete_deletes_Gene_at_specified_or_random_index(self):
        assert len(self.allele.genes) == 4
        self.allele.delete(0)
        assert len(self.allele.genes) == 3
        assert self.allele.genes[0].name == "g2"
        assert self.allele.genes[1].name == "g3"
        assert self.allele.genes[2].name == "g4"

        indices = set()
        for _ in range(10):
            allele = self.first_allele()
            allele.delete()
            if allele.genes[0].name != "g1":
                indices.add(0)
            elif allele.genes[1].name != "g2":
                indices.add(1)
            elif allele.genes[2].name != "g3":
                indices.add(2)
            else:
                indices.add(3)
        assert len(indices) >= 2

    def test_substitute_replaces_Gene_with_specified_or_random_Gene_at_specified_or_random_index(self):
        ref_gene = self.first_gene()
        self.allele.substitute(1, ref_gene)
        assert self.allele.genes[1] == ref_gene

        gs = set()
        indices = set()
        for _ in range(10):
            allele = self.first_allele()
            g_at_1 = allele.genes[1]
            allele.substitute(1)
            assert allele.genes[1] != g_at_1
            gs.add(allele.genes[1])
            allele.substitute(gene=g_at_1)
            indices.add(allele.genes.index(g_at_1))
        assert len(gs) >= 9
        assert len(indices) >= 2

    def test_recombine_swaps_Genes_between_Alleles_at_specified_or_random_indices(self):
        al3 = self.allele.recombine(self.other, [1], False)
        assert al3.genes == [self.allele.genes[0], *self.other.genes[1:]]

        swapset = set()
        for _ in range(10):
            al1 = self.first_allele()
            al2 = self.second_allele()
            al3 = al1.recombine(al2, recombine_genes=False)
            swapped = False
            swaps = []
            for i in range(len(al3.genes)):
                if swapped and al3.genes[i] == al1.genes[i]:
                    swapped = False
                    swaps.append(i)
                elif al3.genes[i] == al2.genes[i]:
                    swapped = True
                    swaps.append(i)
            swapset.add(tuple(swaps))
        assert len(swapset) >= 2

    def test_recombine_recombines_underlying_Genes_with_matching_name_and_index_by_default(self):
        allele = self.allele.recombine(self.alternate, [1])
        recombined_genes = 0
        allele_bases = [b for g in self.allele.genes for b in g.bases]
        other_bases = [b for g in self.alternate.genes for b in g.bases]
        for g in allele.genes:
            if g not in self.allele.genes and g not in self.alternate.genes:
                recombined_genes += 1
            assert all(b in allele_bases or b in other_bases for b in g.bases)
        assert recombined_genes >= 1

    def test_recombine_does_not_recombine_underlying_Genes_with_different_names_by_default(self):
        allele = self.allele.recombine(self.other, [1])
        recombined_genes = 0
        for g in allele.genes:
            if g not in self.allele.genes and g not in self.other.genes:
                recombined_genes += 1
        assert recombined_genes == 0

    def test_recombine_can_recombine_underlying_Genes_with_different_names(self):
        allele = self.allele.recombine(self.other, [1], match_genes=False)
        recombined_genes = 0
        for g in allele.genes:
            if g not in self.allele.genes and g not in self.other.genes:
                recombined_genes += 1
        assert recombined_genes >= 1

    def test_make_returns_random_Allele_with_random_Genes(self):
        allele = genes.Allele.make(3, 5)
        assert type(allele) is genes.Allele
        assert len(allele.genes) == 3
        assert all(len(g.bases) == 5 for g in allele.genes)

        results = set()
        for _ in range(10):
            results.add(genes.Allele.make(5, 10))
        assert len(results) >= 8
        assert len(set([a.name for a in results])) >= 8
        assert len(set([tuple(a.genes) for a in results])) >= 8

    def test_make_with_base_factory_sets_bases_with_correct_type(self):
        float_factory = lambda: random()
        allele = genes.Allele.make(3, 5, base_factory=float_factory)
        assert all(type(b) is float for g in allele.genes for b in g.bases)
        assert all(len(set(tuple(g.bases))) >= 4 for g in allele.genes)

        float_factory = lambda mult: random()*mult
        allele = genes.Allele.make(3, 5, base_factory=float_factory, factory_args=[-2])
        assert all(-2 <= b <= 0 for g in allele.genes for b in g.bases)

    def test_to_dict_and_from_dict_e2e(self):
        encoded = self.allele.to_dict()
        assert type(encoded) is dict
        decoded = genes.Allele.from_dict(encoded)
        assert decoded == self.allele
        assert decoded is not self.allele

        float_factory = lambda: random()
        allele = genes.Allele.make(3, 5, base_factory=float_factory)
        encoded = allele.to_dict()
        assert type(encoded) is dict
        decoded = genes.Allele.from_dict(encoded)
        assert decoded == allele
        assert decoded is not allele

        str_factory = lambda: genes.random_str(5)
        allele = genes.Allele.make(3, 5, base_factory=str_factory)
        encoded = allele.to_dict()
        assert type(encoded) is dict
        decoded = genes.Allele.from_dict(encoded)
        assert decoded == allele
        assert decoded is not allele


class TestChromosome(unittest.TestCase):
    def first_gene(self) -> genes.Gene:
        return genes.Gene("test", [1,2,3])

    def second_gene(self) -> genes.Gene:
        return genes.Gene("diff", [4,5,6])

    def first_allele(self) -> genes.Allele:
        return genes.Allele("test", [
            genes.Gene("g1", [1, 2, 3]),
            genes.Gene("g2", [4, 5, 6]),
            genes.Gene("g3", [7, 8, 9]),
            genes.Gene("g4", [-7, -8, -9]),
        ])

    def alt_allele(self, name: str = "test") -> genes.Allele:
        return genes.Allele(name, [
            genes.Gene("g1", [10, 11, 12]),
            genes.Gene("g2", [13, 14, 15]),
            genes.Gene("g3", [16, 17, 18]),
            genes.Gene("g4", [-16, -17, -18]),
        ])

    def second_allele(self) -> genes.Allele:
        return genes.Allele("diff", [
            genes.Gene("g4", [10, 11, 12]),
            genes.Gene("g5", [13, 14, 15]),
            genes.Gene("g6", [16, 17, 18]),
            genes.Gene("g6", [116, 117, 118]),
        ])

    def third_allele(self) -> genes.Allele:
        return genes.Allele("fizz", [
            genes.Gene("bz1", [20, 21, 22]),
            genes.Gene("bz2", [30, 31, 32]),
            genes.Gene("bz3", [40, 41, 42]),
            genes.Gene("bz4", [140, 141, 142]),
        ])

    def first_chromosome(self) -> genes.Chromosome:
        return genes.Chromosome("test", [
            self.first_allele(),
            self.second_allele()
        ])

    def alt_chromosome(self) -> genes.Chromosome:
        return genes.Chromosome("test", [
            self.alt_allele(),
            self.second_allele()
        ])

    def second_chromosome(self) -> genes.Chromosome:
        return genes.Chromosome("diff", [
            self.second_allele(),
            self.third_allele(),
        ])

    def setUp(self) -> None:
        self.chromosome = self.first_chromosome()
        self.alternate = self.alt_chromosome()
        self.other = self.second_chromosome()
        return super().setUp()

    def test_Chromosome_is_hashable(self):
        hash(self.chromosome)

    def test_copy_returns_identical_Chromosome_containing_identical_Alleles(self):
        copy = self.chromosome.copy()
        assert type(copy) is genes.Chromosome
        assert copy is not self.chromosome
        assert copy == self.chromosome

        for a, b in zip(copy.alleles, self.chromosome.alleles):
            assert a == b
            assert a.name == b.name
            assert a.genes == b.genes
            assert a is not b

    def test_insert_inserts_specified_or_random_allele_at_specified_or_random_index(self):
        al4 = genes.Allele.make(3, 5, "al4")
        al5 = genes.Allele.make(3, 5, "al5")
        self.chromosome.insert(0, al4)
        assert self.chromosome.alleles[0] == al4
        self.chromosome.insert(2, al5)
        assert self.chromosome.alleles[2] == al5

        indices = set()
        ref_allele = self.first_allele().append(self.second_gene())
        for _ in range(10):
            chromosome = self.alt_chromosome()
            chromosome.insert(allele=ref_allele)
            indices.add(chromosome.alleles.index(ref_allele))
        assert len(indices) > 1

        gs = set()
        for _ in range(10):
            chromosome = self.first_chromosome()
            chromosome.insert(0)
            gs.add(chromosome.alleles[0])
        assert len(gs) >= 4

    def test_append_adds_Allele_to_end_of_alleles(self):
        gl1 = genes.Allele("gl1", [self.first_gene()])
        gl2 = genes.Allele("gl2", [self.second_gene()])
        self.chromosome.append(gl1)
        assert self.chromosome.alleles[-1] == gl1
        self.chromosome.append(gl2)
        assert self.chromosome.alleles[-1] == gl2

    def test_duplicate_duplicates_Allele_at_specified_or_random_index(self):
        self.chromosome.duplicate(0)
        assert self.chromosome.alleles[0] == self.chromosome.alleles[1]
        assert self.chromosome.alleles[0] is not self.chromosome.alleles[1]

        indices = set()
        for _ in range(10):
            chromosome = self.first_chromosome().append(self.alt_allele())
            chromosome.duplicate()
            for i in range(len(chromosome.alleles)):
                if chromosome.alleles[i] == chromosome.alleles[i+1]:
                    indices.add(i)
                    break
        assert len(indices) >= 2

    def test_delete_deletes_Allele_at_specified_or_random_index(self):
        assert len(self.chromosome.alleles) == 2
        self.chromosome.delete(0)
        assert len(self.chromosome.alleles) == 1
        assert self.chromosome.alleles[0].name == "diff"

        indices = set()
        for _ in range(10):
            chromosome = self.first_chromosome().append(self.alt_allele("fizz"))
            chromosome.delete()
            if chromosome.alleles[0].name != "test":
                indices.add(0)
            elif chromosome.alleles[1].name != "diff":
                indices.add(1)
            else:
                indices.add(2)
        assert len(indices) >= 2

    def test_substitute_replaces_Allele_with_specified_or_random_Allele_at_specified_or_random_index(self):
        ref_allele = self.first_allele()
        assert self.chromosome.alleles[1] != ref_allele
        self.chromosome.substitute(1, ref_allele)
        assert self.chromosome.alleles[1] == ref_allele

        gs = set()
        indices = set()
        for _ in range(20):
            chromosome = self.first_chromosome()
            a_at_1 = chromosome.alleles[1]
            chromosome.substitute(1)
            assert chromosome.alleles[1] != a_at_1
            gs.add(chromosome.alleles[1])
            chromosome.substitute(allele=a_at_1)
            indices.add(chromosome.alleles.index(a_at_1))
        assert len(gs) >= 9
        assert len(indices) >= 2

    def test_recombine_swaps_Alleles_between_Chromosomes_at_specified_or_random_indices(self):
        al3 = self.chromosome.recombine(self.other, [1], False)
        assert al3.alleles == [self.chromosome.alleles[0], *self.other.alleles[1:]]

        swapset = set()
        for _ in range(10):
            ch1 = self.first_chromosome()
            ch2 = self.second_chromosome()
            ch3 = ch1.recombine(ch2, recombine_genes=False)
            swapped = False
            swaps = []
            for i in range(len(ch3.alleles)):
                if swapped and ch3.alleles[i] == ch1.alleles[i]:
                    swapped = False
                    swaps.append(i)
                elif ch3.alleles[i] == ch2.alleles[i]:
                    swapped = True
                    swaps.append(i)
            swapset.add(tuple(swaps))
        assert len(swapset) >= 2

    def test_recombine_recombines_underlying_Alleles_and_Genes_with_matching_name_and_index_by_default(self):
        chromosome = self.chromosome.recombine(self.alternate, [1])
        main_genes = [g for a in self.chromosome.alleles for g in a.genes]
        main_gene_names = [g.name for g in main_genes]
        main_bases = [b for g in main_genes for b in g.bases]
        other_genes = [b for g in self.alternate.alleles for b in g.genes]
        other_gene_names = [g.name for g in other_genes]
        other_bases = [b for g in other_genes for b in g.bases]
        recombined_alleles = 0
        recombined_genes = 0
        for c in chromosome.alleles:
            if c not in self.chromosome.alleles and c not in self.alternate.alleles:
                recombined_alleles += 1
            for g in c.genes:
                if g not in main_genes and g not in other_genes:
                    recombined_genes += 1
                assert g.name in main_gene_names or g.name in other_gene_names
                assert all(b in main_bases or b in other_bases for b in g.bases)
        assert recombined_alleles >= 1
        assert recombined_genes >= 1

    def test_recombine_does_not_recombine_underlying_Alleles_or_Genes_with_different_names_by_default(self):
        chromosome = self.chromosome.recombine(self.other, [1])
        main_genes = [g for a in self.chromosome.alleles for g in a.genes]
        other_genes = [b for g in self.other.alleles for b in g.genes]
        recombined_alleles = 0
        recombined_genes = 0

        for a in chromosome.alleles:
            if a not in self.chromosome.alleles and a not in self.other.alleles:
                recombined_alleles += 1
            for g in a.genes:
                if g not in main_genes and g not in other_genes:
                    recombined_genes += 1
        assert recombined_alleles == 0
        assert recombined_genes == 0

    def test_recombine_can_recombine_underlying_Alleles_and_Genes_with_different_names(self):
        chromosome = self.chromosome.recombine(
            self.other, [1], match_alleles=False, match_genes=False
        )
        main_genes = [g for a in self.chromosome.alleles for g in a.genes]
        main_gene_names = [g.name for g in main_genes]
        other_genes = [b for g in self.other.alleles for b in g.genes]
        other_gene_names = [g.name for g in other_genes]
        recombined_alleles = 0
        recombined_genes = 0
        for a in chromosome.alleles:
            if a not in self.chromosome.alleles and a not in self.other.alleles:
                recombined_alleles += 1
            for g in a.genes:
                if g.name not in main_gene_names and g.name not in other_gene_names:
                    recombined_genes += 1
        assert recombined_alleles >= 1
        assert recombined_genes >= 1

    def test_make_returns_random_Chromosome_with_random_Alleles_with_random_Genes(self):
        chromosome = genes.Chromosome.make(3, 5, 7)
        assert type(chromosome) is genes.Chromosome
        assert len(chromosome.alleles) == 3
        assert all(len(a.genes) == 5 for a in chromosome.alleles)
        assert all(len(g.bases) == 7 for a in chromosome.alleles for g in a.genes)

        results: set[genes.Chromosome] = set()
        for _ in range(10):
            results.add(genes.Chromosome.make(5, 10, 20))
        assert len(results) >= 8
        assert len(set([c.name for c in results])) >= 8
        assert len(set([tuple(c.alleles) for c in results])) >= 8
        assert len(set([tuple(a.genes) for c in results for a in c.alleles])) >= 8

    def test_make_with_base_factory_sets_bases_with_correct_type(self):
        float_factory = lambda: random()
        chromosome = genes.Chromosome.make(3, 5, 7, base_factory=float_factory)
        assert all(
            type(b) is float
            for a in chromosome.alleles
            for g in a.genes
            for b in g.bases
        )
        assert all(
            len(set(tuple(g.bases))) >= 6
            for a in chromosome.alleles
            for g in a.genes
        )

        float_factory = lambda mult: random()*mult
        chromosome = genes.Chromosome.make(3, 5, 7, base_factory=float_factory, factory_args=[-2])
        assert all(
            -2 <= b <= 0
            for a in chromosome.alleles
            for g in a.genes
            for b in g.bases
        )

    def test_to_dict_and_from_dict_e2e(self):
        encoded = self.chromosome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Chromosome.from_dict(encoded)
        assert decoded == self.chromosome
        assert decoded is not self.chromosome

        float_factory = lambda: random()
        chromosome = genes.Chromosome.make(3, 5, 7, base_factory=float_factory)
        encoded = chromosome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Chromosome.from_dict(encoded)
        assert decoded == chromosome
        assert decoded is not chromosome

        str_factory = lambda: genes.random_str(5)
        chromosome = genes.Chromosome.make(3, 5, 7, base_factory=str_factory)
        encoded = chromosome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Chromosome.from_dict(encoded)
        assert decoded == chromosome
        assert decoded is not chromosome


class TestGnome(unittest.TestCase):
    def first_gene(self) -> genes.Gene:
        return genes.Gene("test", [1,2,3])

    def second_gene(self) -> genes.Gene:
        return genes.Gene("diff", [4,5,6])

    def first_allele(self) -> genes.Allele:
        return genes.Allele("test", [
            genes.Gene("g1", [1, 2, 3]),
            genes.Gene("g2", [4, 5, 6]),
            genes.Gene("g3", [7, 8, 9]),
            genes.Gene("g4", [-7, -8, -9]),
        ])

    def alt_allele(self, name: str = "test") -> genes.Allele:
        return genes.Allele(name, [
            genes.Gene("g1", [10, 11, 12]),
            genes.Gene("g2", [13, 14, 15]),
            genes.Gene("g3", [16, 17, 18]),
            genes.Gene("g4", [-16, -17, -18]),
        ])

    def second_allele(self) -> genes.Allele:
        return genes.Allele("diff", [
            genes.Gene("g4", [10, 11, 12]),
            genes.Gene("g5", [13, 14, 15]),
            genes.Gene("g6", [16, 17, 18]),
            genes.Gene("g6", [116, 117, 118]),
        ])

    def third_allele(self) -> genes.Allele:
        return genes.Allele("fizz", [
            genes.Gene("bz1", [20, 21, 22]),
            genes.Gene("bz2", [30, 31, 32]),
            genes.Gene("bz3", [40, 41, 42]),
            genes.Gene("bz4", [140, 141, 142]),
        ])

    def fourth_allele(self) -> genes.Allele:
        return genes.Allele("buzz", [
            genes.Gene("bz4", [220, 221, 222]),
            genes.Gene("bz5", [230, 231, 232]),
            genes.Gene("bz6", [240, 241, 242]),
            genes.Gene("bz7", [250, 251, 252]),
        ])

    def first_chromosome(self) -> genes.Chromosome:
        return genes.Chromosome("test", [
            self.first_allele(),
            self.second_allele()
        ])

    def alt_chromosome(self, name: str = "test") -> genes.Chromosome:
        return genes.Chromosome(name, [
            self.alt_allele(),
            self.second_allele()
        ])

    def second_chromosome(self, name: str = "diff") -> genes.Chromosome:
        return genes.Chromosome(name, [
            self.second_allele(),
            self.third_allele(),
        ])

    def third_chromosome(self) -> genes.Chromosome:
        return genes.Chromosome("fizz", [
            self.third_allele(),
            self.fourth_allele(),
        ])

    def first_genome(self) -> genes.Genome:
        return genes.Genome("test", [
            self.first_chromosome(),
            self.second_chromosome(),
        ])

    def alt_genome(self) -> genes.Genome:
        return genes.Genome("test", [
            self.alt_chromosome(),
            self.second_chromosome(),
        ])

    def second_genome(self) -> genes.Genome:
        return genes.Genome("diff", [
            self.second_chromosome("asdf"),
            self.third_chromosome(),
        ])

    def setUp(self) -> None:
        self.genome = self.first_genome()
        self.alternate = self.alt_genome()
        self.other = self.second_genome()
        return super().setUp()

    def test_Genome_is_hashable(self):
        hash(self.genome)

    def test_copy_returns_identical_Genome_containing_identical_Chromosomes(self):
        copy = self.genome.copy()
        assert type(copy) is genes.Genome
        assert copy is not self.genome
        assert copy == self.genome

        for a, b in zip(copy.chromosomes, self.genome.chromosomes):
            assert a == b
            assert a.name == b.name
            assert a.alleles == b.alleles
            assert a is not b

    def test_insert_inserts_specified_or_random_Chromosome_at_specified_or_random_index(self):
        ch4 = genes.Chromosome.make(3, 5, 7, "ch4")
        ch5 = genes.Chromosome.make(3, 5, 7, "ch5")
        self.genome.insert(0, ch4)
        assert self.genome.chromosomes[0] == ch4
        self.genome.insert(2, ch5)
        assert self.genome.chromosomes[2] == ch5

        indices = set()
        ref_chromosome = self.first_chromosome()
        for _ in range(10):
            genome = self.alt_genome().append(self.third_chromosome())
            genome.insert(chromosome=ref_chromosome)
            indices.add(genome.chromosomes.index(ref_chromosome))
        assert len(indices) > 1

        gs = set()
        for _ in range(10):
            genome = self.first_genome()
            genome.insert(0)
            gs.add(genome.chromosomes[0])
        assert len(gs) >= 4

    def test_append_adds_Chromosome_to_end_of_chromosomes(self):
        ch1 = genes.Chromosome("ch1", [self.first_chromosome()])
        ch2 = genes.Chromosome("ch2", [self.second_chromosome()])
        self.genome.append(ch1)
        assert self.genome.chromosomes[-1] == ch1
        self.genome.append(ch2)
        assert self.genome.chromosomes[-1] == ch2

    def test_duplicate_duplicates_Chromosome_at_specified_or_random_index(self):
        self.genome.duplicate(0)
        assert self.genome.chromosomes[0] == self.genome.chromosomes[1]
        assert self.genome.chromosomes[0] is not self.genome.chromosomes[1]

        indices = set()
        for _ in range(10):
            genome = self.first_genome().append(self.alt_chromosome())
            genome.duplicate()
            for i in range(len(genome.chromosomes)):
                if genome.chromosomes[i] == genome.chromosomes[i+1]:
                    indices.add(i)
                    break
        assert len(indices) >= 2

    def test_delete_deletes_Chromosome_at_specified_or_random_index(self):
        assert len(self.genome.chromosomes) == 2
        self.genome.delete(0)
        assert len(self.genome.chromosomes) == 1
        assert self.genome.chromosomes[0].name == "diff"

        indices = set()
        for _ in range(10):
            genome = self.first_genome().append(self.alt_chromosome("fizz"))
            genome.delete()
            if genome.chromosomes[0].name != "test":
                indices.add(0)
            elif genome.chromosomes[1].name != "diff":
                indices.add(1)
            else:
                indices.add(2)
        assert len(indices) >= 2

    def test_substitute_replaces_Chromosome_with_specified_or_random_Chromosome_at_specified_or_random_index(self):
        ref_chromosome = self.first_chromosome()
        assert self.genome.chromosomes[1] != ref_chromosome
        self.genome.substitute(1, ref_chromosome)
        assert self.genome.chromosomes[1] == ref_chromosome

        gs = set()
        indices = set()
        for _ in range(10):
            genome = self.first_genome().append(self.third_chromosome())
            c_at_1 = genome.chromosomes[1]
            genome.substitute(1)
            assert genome.chromosomes[1] != c_at_1
            gs.add(genome.chromosomes[1])
            genome.substitute(chromosome=c_at_1)
            indices.add(genome.chromosomes.index(c_at_1))
        assert len(gs) >= 9
        assert len(indices) >= 2

    def test_recombine_swaps_Chromosomes_between_Genomes_at_specified_or_random_indices(self):
        gm3 = self.genome.recombine(self.other, [1], False)
        assert gm3.chromosomes == [self.genome.chromosomes[0], *self.other.chromosomes[1:]]

        swapset = set()
        for _ in range(10):
            gm1 = self.first_genome()
            gm2 = self.second_genome()
            gm3 = gm1.recombine(gm2, recombine_alleles=False)
            swapped = False
            swaps = []
            for i in range(len(gm3.chromosomes)):
                if swapped and gm3.chromosomes[i] == gm1.chromosomes[i]:
                    swapped = False
                    swaps.append(i)
                elif gm3.chromosomes[i] == gm2.chromosomes[i]:
                    swapped = True
                    swaps.append(i)
            swapset.add(tuple(swaps))
        assert len(swapset) >= 2

    def test_recombine_recombines_underlying_Chromosomes_Alleles_and_Genes_with_matching_name_and_index_by_default(self):
        genome = self.genome.recombine(self.alternate, [1])
        main_alleles = [a for c in self.genome.chromosomes for a in c.alleles]
        main_genes = [g for a in main_alleles for g in a.genes]
        main_gene_names = [g.name for g in main_genes]
        main_bases = [b for g in main_genes for b in g.bases]
        other_alleles = [a for c in self.alternate.chromosomes for a in c.alleles]
        other_genes = [g for a in other_alleles for g in a.genes]
        other_gene_names = [g.name for g in other_genes]
        other_bases = [b for g in other_genes for b in g.bases]
        recombined_chromosomes = 0
        recombined_alleles = 0
        recombined_genes = 0
        for c in genome.chromosomes:
            if c not in self.genome.chromosomes and c not in self.alternate.chromosomes:
                recombined_chromosomes += 1
            for a in c.alleles:
                if a not in main_alleles and a not in other_alleles:
                    recombined_alleles += 1
                for g in a.genes:
                    if g not in main_genes and g not in other_genes:
                        recombined_genes += 1
                    assert g.name in main_gene_names or g.name in other_gene_names
                    assert all(b in main_bases or b in other_bases for b in g.bases)
        assert recombined_chromosomes >= 1
        assert recombined_alleles >= 1
        assert recombined_genes >= 1

    def test_recombine_does_not_recombine_underlying_Chromosomes_or_Genes_with_different_names_by_default(self):
        genome = self.genome.recombine(self.other, [1])
        main_alleles = [a for c in self.genome.chromosomes for a in c.alleles]
        main_genes = [g for a in main_alleles for g in a.genes]
        other_alleles = [a for c in self.other.chromosomes for a in c.alleles]
        other_genes = [g for a in other_alleles for g in a.genes]
        recombined_chromosomes = 0
        recombined_alleles = 0
        recombined_genes = 0

        for c in genome.chromosomes:
            if c not in self.genome.chromosomes and c not in self.other.chromosomes:
                recombined_chromosomes += 1
            for a in c.alleles:
                if a not in main_alleles and a not in other_alleles:
                    recombined_alleles += 1
                for g in a.genes:
                    if g not in main_genes and g not in other_genes:
                        recombined_genes += 1
        assert recombined_chromosomes == 0
        assert recombined_alleles == 0
        assert recombined_genes == 0

    def test_recombine_can_recombine_underlying_Chromosomes_Alleles_and_Genes_with_different_names(self):
        genome = self.genome.recombine(
            self.other, [1], match_chromosomes=False, match_alleles=False,
            match_genes=False
        )
        main_alleles = [a for c in self.genome.chromosomes for a in c.alleles]
        main_genes = [g for a in main_alleles for g in a.genes]
        other_alleles = [a for c in self.other.chromosomes for a in c.alleles]
        other_genes = [g for a in other_alleles for g in a.genes]
        recombined_chromosomes = 0
        recombined_alleles = 0
        recombined_genes = 0
        for c in genome.chromosomes:
            if c not in self.genome.chromosomes and c not in self.other.chromosomes:
                recombined_chromosomes += 1
            for a in c.alleles:
                if a not in main_alleles and a not in other_alleles:
                    recombined_alleles += 1
                for g in a.genes:
                    if g not in main_genes and g not in other_genes:
                        recombined_genes += 1
        assert recombined_chromosomes >= 1
        assert recombined_alleles >= 1
        assert recombined_genes >= 1

    def test_make_returns_random_Genome_with_random_Chromosomes_Alleles_and_Genes(self):
        genome = genes.Genome.make(2, 3, 5, 7)
        assert type(genome) is genes.Genome
        assert len(genome.chromosomes) == 2
        assert all(len(c.alleles) == 3 for c in genome.chromosomes)
        assert all(len(a.genes) == 5 for c in genome.chromosomes for a in c.alleles)
        assert all(
            len(g.bases) == 7
            for c in genome.chromosomes
            for a in c.alleles
            for g in a.genes
        )

        results: set[genes.Genome] = set()
        for _ in range(10):
            results.add(genes.Genome.make(3, 5, 10, 20))
        assert len(results) >= 8
        assert len(set([n.name for n in results])) >= 8
        assert len(set([tuple(n.chromosomes) for n in results])) >= 8
        assert len(set([tuple(c.alleles) for n in results for c in n.chromosomes])) >= 8

    def test_make_with_base_factory_sets_bases_with_correct_type(self):
        float_factory = lambda: random()
        genome = genes.Genome.make(2, 3, 5, 7, base_factory=float_factory)
        assert all(
            type(b) is float
            for c in genome.chromosomes
            for a in c.alleles
            for g in a.genes
            for b in g.bases
        )
        assert all(
            len(set(tuple(g.bases))) >= 6
            for c in genome.chromosomes
            for a in c.alleles
            for g in a.genes
        )

        float_factory = lambda mult: random()*mult
        genome = genes.Genome.make(2, 3, 5, 7, base_factory=float_factory, factory_args=[-2])
        assert all(
            -2 <= b <= 0
            for c in genome.chromosomes
            for a in c.alleles
            for g in a.genes
            for b in g.bases
        )

    def test_to_dict_and_from_dict_e2e(self):
        encoded = self.genome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Genome.from_dict(encoded)
        assert decoded == self.genome
        assert decoded is not self.genome

        float_factory = lambda: random()
        genome = genes.Genome.make(2, 3, 5, 7, base_factory=float_factory)
        encoded = genome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Genome.from_dict(encoded)
        assert decoded == genome
        assert decoded is not genome

        str_factory = lambda: genes.random_str(5)
        genome = genes.Genome.make(2, 3, 5, 7, base_factory=str_factory)
        encoded = genome.to_dict()
        assert type(encoded) is dict
        decoded = genes.Genome.from_dict(encoded)
        assert decoded == genome
        assert decoded is not genome


if __name__ == '__main__':
    unittest.main()
